# vim:fileencoding=utf-8
# Copyright (c) 2008-2009 gocept gmbh & co. kg
# See also LICENSE.txt
# $Id$
"""Munin data views."""

import gocept.lms.app
import gocept.lms.interfaces
import grok
import zc.queue.interfaces
import zope.component


grok.context(gocept.lms.app.LMS)


class ViewMuninData(grok.Permission):
    grok.name('gocept.lms.ViewMuninData')


class StatisticalView(object):
    grok.require(ViewMuninData)

    @property
    def stats(self):
        return zope.component.getUtility(gocept.lms.interfaces.IStatistics)

class QueueSize(StatisticalView, grok.View):
    grok.name('munin-queuesize')
    grok.require(ViewMuninData)

    def render(self):
        return 'queue-size:%s' % self.stats.queue_size


class ThreadPool(StatisticalView, grok.View):
    grok.name('munin-threadpool')
    grok.require(ViewMuninData)

    def render(self):
        return 'thread-pool:%s' % self.stats.thread_pool


class Statistics(grok.Model):

    zope.interface.implements(gocept.lms.interfaces.IStatistics)
    
    thread_pool = 0

    @property
    def queue_size(self):
        check_queue = zope.component.getUtility(
                zc.queue.interfaces.IQueue, name='check')
        return len(check_queue)


@grok.subscribe(gocept.lms.app.LMS, grok.IObjectAddedEvent)
def add_statistics(lms, event):
    sm = lms.getSiteManager()
    sm['statistics'] = Statistics()
    sm.registerUtility(sm['statistics'],
        provided=gocept.lms.interfaces.IStatistics)



# Copyright (c) 2008-2009 gocept gmbh & co. kg
# See also LICENSE.txt

import logging
import os
import pprint
import decorator

import transaction
import BTrees
import grok
import zc.catalog.catalogindex
import zc.queue
import zc.queue.interfaces
import zope.app.catalog.catalog
import zope.app.catalog.field
import zope.app.catalog.interfaces
import zope.app.intid
import zope.app.intid.interfaces
import zope.component
import zope.security.interfaces

import gocept.lms.interfaces
import gocept.lms.url
import gocept.lms.client

log = logging.getLogger(__name__)

PROTOCOL_VERSION = 2


def setup_url_catalog(catalog):
    catalog['state'] = zope.app.catalog.field.FieldIndex(
        'state', gocept.lms.interfaces.IURL)
    catalog['last_check'] = zope.app.catalog.field.FieldIndex(
        'last_check', gocept.lms.interfaces.IURL)
    catalog['last_state_change'] = zope.app.catalog.field.FieldIndex(
        'last_state_change', gocept.lms.interfaces.IURL)
    catalog['clients'] = zc.catalog.catalogindex.SetIndex(
        'clients', gocept.lms.interfaces.IURL)


class LMS(grok.Application, grok.Container):

    grok.local_utility(gocept.lms.url.URLContainer,
                       provides=gocept.lms.interfaces.IURLProvider,
                       public=True)

    grok.local_utility(gocept.lms.client.ClientContainer,
                       provides=gocept.lms.interfaces.IClientProvider,
                       public=True)

    grok.local_utility(zope.app.intid.IntIds,
                       provides=zope.app.intid.interfaces.IIntIds)

    grok.local_utility(zope.app.catalog.catalog.Catalog,
                       provides=zope.app.catalog.interfaces.ICatalog,
                       setup=setup_url_catalog,
                       name='urls')

    grok.local_utility(zc.queue.CompositeQueue,
                       provides=zc.queue.interfaces.IQueue,
                       name='check')


class Index(grok.View):
    pass # see app_templates/index.pt


class Alerts(grok.View):
    pass # see app_templates/alerts.pt


class URLs(grok.View):

    def update(self, pattern=None):
        self.pattern = pattern

    def urls(self):
        if not self.pattern:
            return
        urls = zope.component.getUtility(gocept.lms.interfaces.IURLProvider)
        for url in urls.values():
            if self.pattern in url.url:
                yield url


@decorator.decorator
def authenticatedClient(method, self, client_id, password, *args, **kwargs):
    """Make sure the client is authenticated and pass the client object to
    method.

    """
    clients = zope.component.getUtility(
        gocept.lms.interfaces.IClientProvider)
    client = clients.get(client_id)
    if client is None or client.password != password:
        raise zope.security.interfaces.Unauthorized
    return method(self, client, password, *args, **kwargs)

@decorator.decorator
def logged(method, self, *args):
    args_str = pprint.pformat(args)
    note = (u'XML-RPC: %s.%s\n%s' % (
        self.__class__.__name__, method.__name__, args_str))[:1000]
    transaction.get().note(note.encode('utf8'))
    log.info('%s: %s' % (self.request.principal.title, note.encode('utf-8')))
    result = method(self, *args)
    if not transaction.get().description:
        # if transaction note got lost, set it here again
        transaction.get().note(note.encode('utf8'))
    note = note + u'\nResult: %s' % result
    log.debug('%s: %s' % (self.request.principal.title, note.encode('utf-8')))
    return result


class XMLRPC(grok.XMLRPC):

    def registerClient(self, secret, id, callback,
                       contact_name, contact_email):
        client = gocept.lms.client.Client(id)
        client.callback = callback
        client.contact_name = contact_name
        client.contact_email = contact_email
        client.password = os.urandom(24).encode('base64')[:-3]
        clients = zope.component.getUtility(
            gocept.lms.interfaces.IClientProvider)
        clients.add(client)
        return True

    @logged
    @authenticatedClient
    def checkConnection(self, client, password):
        return PROTOCOL_VERSION

    @logged
    @authenticatedClient
    def getClientNotifications(self, client, password):
        return gocept.lms.interfaces.INotifications(client).enabled

    @logged
    @authenticatedClient
    def setClientNotifications(self, client, password, status):
        gocept.lms.interfaces.INotifications(client).enabled = bool(status)

    @logged
    @authenticatedClient
    def registerURLs(self, client, password, urls):
        url_objects = list(self.url_provider.add_multiple(urls))
        client.register_urls(url_objects)
        result = [(url.url, url.state, url.reason)
                  for url in url_objects
                  if url.state is not None]
        return result

    @logged
    @authenticatedClient
    def unregisterURLs(self, client, password, urls):
        url_objects = (url for url in
                       (self.url_provider.get_url(url) for url in urls)
                       if url is not None)
        client.unregister_urls(url_objects)
        return True

    @logged
    @authenticatedClient
    def forceSynchronization(self, client, password):
        sync = gocept.lms.interfaces.ISynchronization(client)
        sync.force = True

    @logged
    def getInfoFrameURL(self, client_id, password):
        # XXX API stub
        return 'http://lms.gocept.com/v2/alerts'

    # XXX missing
    # listClients(admin_secret)


    @property
    def url_provider(self):
        return zope.component.getUtility(gocept.lms.interfaces.IURLProvider)



# Copyright (c) 2008-2009 gocept gmbh & co. kg
# See also LICENSE.txt

import logging

import grok
import zope.app.appsetup.product
import zope.app.container.interfaces
import zope.component
import zope.sendmail.interfaces

import gocept.lms.interfaces


log = logging.getLogger(__name__)


# XXX we actually would like to get the texts out of here

# When a client gets disabled, we send a mail:
MAIL_DISABLED = """\
To: %(client_name)s <%(client_email)s>
From: Link Monitoring Server <%(lms_email)s>
Subject: lms disabled notifications for client id %(client_id)s

Dear %(client_name)s

Your site is registered with our Link Monitoring Server (lms) at %(lms_name)s.

The lms will not send you any further notifications, because it could not send
out notifications for several times. To enable notifications again, visit your
Link Management configuration (from CMFLinkChecker 1.9.2 on).

Information stored on the lms:

    Client-Id   : %(client_id)s
    Callback URL: %(client_callback)s
    Contact     : %(client_name)s <%(client_email)s>

"""

# When a client is registered we send this mail:
MAIL_REGISTRATION = """\
To: %(client_name)s <%(client_email)s>
From: Link Monitoring Server <%(lms_email)s>
Subject: Registration of client id "%(client_id)s"

Dear %(client_name)s

Your client id "%(client_id)s" has been registered with the lms %(lms_name)s.

Thanks for trying out CMFLinkChecker and the Link Monitoring Server. If you
have not done so already, please subscribe to the CMFLinkChecker and lms
mailing list at linkchecker@lists.gocept.com (with subscribe as subject).

If you have any problems using CMFLinkChecker feel free to contact us at
mail@gocept.com or the CMFLinkChecker mailing list.

Information stored on the lms:

    Client-Id   : %(client_id)s
    Password    : %(client_password)s
    Callback URL: %(client_callback)s
    Contact     : %(client_name)s <%(client_email)s>

The address of the lms is <%(lms_address)s>.

"""

# When a client is registered we send this mail:
MAIL_UNREGISTRATION = """\
To: %(client_name)s <%(client_email)s>
From: Link Monitoring Server <%(lms_email)s>
Subject: Removal of client id "%(client_id)s"

Dear %(client_name)s

Your client id "%(client_id)s" was registered with the lms
%(lms_name)s. Due to a request for removal, it was now removed from
our database again.

We thank you for using CMFLinkChecker and the Link Monitoring Server and hope
to hear from you in the future again.

If you have further questions feel free to contact us at mail@gocept.com or the
CMFLinkChecker mailing list.

Information removed from the lms:

    Client-Id   : %(client_id)s
    Callback URL: %(client_callback)s
    Contact     : %(client_name)s <%(client_email)s>

Please note that you still might receive notifications or link checks
for another few hours.

"""


def send_mail_to_client(client, body_template):
    mailer = zope.component.queryUtility(
        zope.sendmail.interfaces.IMailDelivery)
    if mailer is None:
        log.error("Could not send mail because not mail delivery utility "
                  "is registered.")
        return

    config = zope.app.appsetup.product.getProductConfiguration('gocept.lms')
    body = body_template % dict(
        client_id=client.id,
        client_password=client.password,
        client_callback=client.callback,
        client_name=client.contact_name,
        client_email=client.contact_email,
        lms_email=config['lms-email-address'],
        lms_name=config['lms-name'],
        lms_address=config['lms-xmlrpc-address'],
    )

    rcpt = [client.contact_email]
    cc = config.get('lms-send-cc-to')
    if cc:
        rcpt.extend(cc.split())

    mailer.send(config['lms-email-address'], rcpt, body)


@grok.subscribe(
    gocept.lms.interfaces.IClient,
    zope.app.container.interfaces.IObjectAddedEvent)
def on_registration(client, event):
    send_mail_to_client(client, MAIL_REGISTRATION)



# Copyright (c) 2008-2009 gocept gmbh & co. kg
# See also LICENSE.txt
"""Perform daily synchronization of client databases."""

import datetime
import gocept.lms.interfaces
import grok
import logging
import pytz
import zope.component
import transaction

log = logging.getLogger(__name__)

INTERVAL = datetime.timedelta(seconds=24*3600)


class Synchronization(grok.Annotation, grok.Model):

    grok.provides(gocept.lms.interfaces.ISynchronization)
    grok.context(gocept.lms.interfaces.IClient)

    last = datetime.datetime.min.replace(tzinfo=pytz.UTC)
    force = False

    @property
    def client(self):
        return self.__parent__

    def sync(self):
        log.info('%s: performing synchronization' % self.client.id)
        log.info('%s: %s URLs currently registered' % (
            self.client.id, len(self.client.urls)))

        try:
            remote_urls = self.client.remote_urls()
        except Exception, e:
            log.exception(e)
            now = datetime.datetime.now(pytz.UTC)
            self.last = now
            self.force = False
            return

        # Make unique: we've seen clients submit the same URL multiple times
        # in one request.
        remote_urls = set(remote_urls)

        url_provider = zope.component.getUtility(gocept.lms.interfaces.IURLProvider)
        remote_urls = set(url_provider.add_multiple(remote_urls))

        log.info('%s: retrieved %s URLs' % (self.client.id, len(remote_urls)))

        local_urls = set(self.client.urls)

        # Step 1: Remove all URLs not known to the client anymore
        self.client.unregister_urls(local_urls-remote_urls)

        # Step 2: Add URLs that appeared at the client without registration
        self.client.register_urls(remote_urls-local_urls)
        log.info('%s: %s URLs remaining' % (
            self.client.id, len(self.client.urls)))

        # Step 3: Notify client about the status of all his URLs
        try:
            self.client.notify_all()
        except Exception, e:
            log.exception(e)
            now = datetime.datetime.now(pytz.UTC)
            self.last = now
            self.force = False
            return

        # Mark as updated
        notifications = gocept.lms.interfaces.INotifications(self.client)
        now = datetime.datetime.now(pytz.UTC)
        notifications.last = self.last = now
        self.force = False


def sync():
    reference_time = datetime.datetime.now(pytz.UTC) - INTERVAL

    clients = zope.component.getUtility(gocept.lms.interfaces.IClientProvider)
    for client in clients.values():
        notifications = gocept.lms.interfaces.INotifications(client)
        if not notifications.enabled:
            continue
        client_sync = gocept.lms.interfaces.ISynchronization(client)
        if client_sync.last > reference_time and not client_sync.force:
            continue
        client_sync.sync()
        transaction.commit()



# Copyright (c) 2008-2009 gocept gmbh & co. kg
# See also LICENSE.txt

import gocept.lms.check
import gocept.lms.notify
import gocept.lms.schedule
import gocept.lms.sync
import gocept.runner


@gocept.runner.appmain(ticks=60)
def scheduler():
    return gocept.lms.schedule.schedule()


@gocept.runner.appmain(ticks=1)
def checker():
    return gocept.lms.check.check_and_process()


@gocept.runner.appmain(ticks=1)
def notifier():
    return gocept.lms.notify.notify()


@gocept.runner.appmain(ticks=60)
def syncer():
    return gocept.lms.sync.sync()



# Copyright (c) 2008-2009 gocept gmbh & co. kg
# See also LICENSE.txt

import transaction
import zope.app.component.hooks
import zope.component

import gocept.lms.interfaces


lms = root['lms']
zope.app.component.hooks.setSite(lms)

bad_urls = set()

urls = zope.component.getUtility(gocept.lms.interfaces.IURLProvider)
for key, url in urls.items():
    handler = gocept.lms.interfaces.ISchemeHandler(url.url, None)
    if not(handler and handler.allow(url.url)):
        bad_urls.add(url.url)

for url in bad_urls:
    transaction.begin()
    urls = zope.component.getUtility(gocept.lms.interfaces.IURLProvider)
    url = urls.get_url(url)
    print url.url
    for client in url.clients:
        client.unregister_urls([url])
    del url.__parent__[url.__name__]
    transaction.commit()



# vim:fileencoding=utf-8
# Copyright (c) 2008-2009 gocept gmbh & co. kg
# See also LICENSE.txt
# $Id$
"""Checker implementation for the FTP."""

import logging
import urllib
import ftplib
import socket
import grok

import gocept.lms.handler

from gocept.lms.interfaces import STATE_OK, STATE_UNAVAILABLE, STATE_TEMPORARY

log = logging.getLogger(__name__)


class FTP(gocept.lms.handler.AbstractURLHandler,
          grok.GlobalUtility):

    grok.name('ftp')

    def check(self, url):
        protocol, url = urllib.splittype(url)
        host, path = urllib.splithost(url)
        username, host = urllib.splituser(host)
        host, port = urllib.splitport(host)

        if username:
            username, password = urllib.splitpasswd(username)
        else:
            username = 'anonymous'
            password = 'mail+lms@gocept.com'

        try:
            connection = ftplib.FTP(host, username, password)
        except socket.gaierror, e:
            return STATE_UNAVAILABLE, e.args[1]
        except Exception, e:
            return STATE_UNAVAILABLE, str(e)

        # Try to find out whether this is a file or a directory
        try:
            connection.set_pasv(True)
            lst = connection.nlst(path)
        except Exception, e:
            connection.quit()
            return STATE_UNAVAILABLE, str(e)

        if lst == [path]:
            # `path` referes to a file. Try to download a bit.
            try:
                connection.retrbinary('RETR %s' % path, accept_data)
            except FTPSuccess:
                connection.close()
                return STATE_OK, 'Transfer successfully started.'
            except Exception, e:
                connection.quit()
                return STATE_UNAVAILABLE, str(e)
        else:
            # `path` refers to a directory and we were able to list it.
            connection.quit()
            return STATE_OK, 'Directory successfully listed'


class FTPSuccess(Exception):
    """Signal that FTP data transfer started."""

def accept_data(data):
    raise FTPSuccess()



# Copyright (c) 2008-2009 gocept gmbh & co. kg
# See also LICENSE.txt

import transaction
import zope.app.component.hooks
import zc.queue
import zc.queue.interfaces


lms = root['lms']
zope.app.component.hooks.setSite(lms)

print "Re-initializing checking queue"

site_manager = lms.getSiteManager()
del site_manager['CompositeQueue']
site_manager['CompositeQueue'] = queue = zc.queue.CompositeQueue()
site_manager.registerUtility(queue, zc.queue.interfaces.IQueue, name='check')

transaction.commit()



# -*- coding: latin-1 -*-
# Copyright (c) 2009 gocept gmbh & co. kg
# See also LICENSE.txt
# $Id: __init__.py 1726 2009-02-03 13:26:25Z ctheune $

import zope.app.generations.generations

SchemaManager = zope.app.generations.generations.SchemaManager(
    minimum_generation=1, generation=1,package_name=__name__)



# Copyright (c) 2008-2009 gocept gmbh & co. kg
# See also LICENSE.txt

import BaseHTTPServer
import SimpleHTTPServer
import gocept.lms
import gocept.lms.client
import os.path
import socket
import unittest
import zope.app.appsetup.product
import zope.app.testing.functional
import zope.sendmail.interfaces
import zope.testing.doctest


ftesting_zcml = os.path.join(
    os.path.dirname(gocept.lms.__file__), 'ftesting.zcml')
FunctionalLayer = zope.app.testing.functional.ZCMLLayer(
    ftesting_zcml, __name__, 'FunctionalLayer', allow_teardown=True)


flags = (zope.testing.doctest.ELLIPSIS |
         zope.testing.doctest.NORMALIZE_WHITESPACE |
         zope.testing.doctest.REPORT_NDIFF |
         zope.testing.doctest.INTERPRET_FOOTNOTES)


class TestHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
    def log_message(self, format, *args):
        # Silence, please.
        pass

    def do_GET(self):
        if self.path == '/agent':
            self.send_response(599, self.headers['user-agent'])
            self.end_headers()
            return ''
        return SimpleHTTPServer.SimpleHTTPRequestHandler.do_GET(self)


class DummyMailDelivery(object):

    zope.interface.implements(zope.sendmail.interfaces.IMailDelivery)

    messages = []

    def send(self, fromaddr, toaddr, msg):
        print "(%s -> %s)" % (fromaddr, toaddr)
        is_base64 = False
        lines = iter(msg.splitlines(True))
        headers = []
        for line in lines:
            print line,
            headers.append(line)
            if not line.strip():
                break
            if line.strip() == "Content-Transfer-Encoding: base64":
                is_base64 = True
        body = "".join(lines)
        print body
        self.messages.append((fromaddr, toaddr, headers, body))

    @classmethod
    def reset(cls):
        cls.messages = []


zope.testing.cleanup.addCleanUp(DummyMailDelivery.reset)


class XMLRPCDummy(object):

    fail_update = False

    def __init__(self):
        self.log = []

    def __call__(self, url):
        self.log.append('Connect: %s' % url)
        return self

    def updateManyStates(self, id, password, notifications):
        if self.fail_update:
            raise Exception('Client unreachable')
        self.log.append('Update many states: %s - %s' %
          (id, len(notifications)))

    def show_log(self):
        print '\n'.join(self.log)
        self.log = []

    def xmlrpc_getAllLinks(self, id, password):
        self.log.append('Get all links')
        return ['http://example.com/1',
                'http://example.com/3']


def install_xmlrpcdummy():
    gocept.lms.client.ServerProxy = dummy = XMLRPCDummy()
    return dummy


def setup(test):
    config = zope.app.appsetup.product._configs['gocept.lms'] = {}
    config['lms-name'] = 'Test LMS'
    config['lms-email-address'] = 'lms@gocept.testing'
    config['lms-xmlrpc-address'] = 'http://lms.testing/'
    config['lms-send-cc-to'] = 'admin@gocept.testing otheradmin@gocept.testing'

def test_suite():
    functional = zope.app.testing.functional.FunctionalDocFileSuite(
        'app.txt',
        'url.txt',
        'check.txt',
        'ftp.txt',
        'http.txt',
        'munin.txt',
        'notify.txt',
        'runner.txt',
        'schedule.txt',
        'sync.txt',
        'urlhandler.txt',
        'xmlrpc.txt',
        optionflags=flags,
        setUp=setup)
    functional.layer = FunctionalLayer
    return unittest.TestSuite([functional])



# vim:fileencoding=utf-8
# Copyright (c) 2008-2009 gocept gmbh & co. kg
# See also LICENSE.txt
# $Id: mailto.py 30834 2010-07-02 10:45:20Z nilo $
"""Checker implementation for mailto: links."""

from gocept.lms.interfaces import STATE_OK, STATE_UNAVAILABLE
import gocept.lms.handler
import grok
import re
import urllib


class Mailto(gocept.lms.handler.AbstractURLHandler,
           grok.GlobalUtility):

    grok.name('mailto')

    def check(self, url):
        mailto, value = urllib.splittype(url)
        if re.match("^[a-zA-Z0-9._%-]+@[a-zA-Z0-9._%-]+.[a-zA-Z]{2,6}$", value):
            status = STATE_OK
            status_msg = 'Ok'
        else:
            status = STATE_UNAVAILABLE
            status_msg = 'No valid email address.'
        return status, status_msg

    def allow(self, url):
        return True

    def classify(self, url):
        return url



# Copyright (c) 2008-2009 gocept gmbh & co. kg
# See also LICENSE.txt

import datetime
import pytz
import zope.component
import gocept.lms.interfaces
import logging

INTERVAL = datetime.timedelta(seconds=15)

log = logging.getLogger(__name__)


def notify():
    reference_time = datetime.datetime.now(pytz.UTC) - INTERVAL

    # Notify each client about URL changes that happened since the last
    # notification.
    clients = zope.component.getUtility(gocept.lms.interfaces.IClientProvider)
    for client in clients.values():
        notifications = gocept.lms.interfaces.INotifications(client)
        if not notifications.enabled:
            continue
        if notifications.last > reference_time:
            # don't notify clients more frequently than every 5 minutes
            continue
        notifications.notify()



# Copyright (c) 2008-2009 gocept gmbh & co. kg
# See also LICENSE.txt

import datetime
import xmlrpclib
import logging

import pytz
import BTrees
import grok
import grokcore.formlib.formlib
from hurry.query import Ge
from hurry.query.set import AnyOf
import hurry.query.interfaces
import zope.interface
import zope.component
import zope.event
import zope.lifecycleevent
import gocept.reference

import gocept.lms.interfaces

log = logging.getLogger(__name__)

# Provide hook for tests to set up dummies
ServerProxy = xmlrpclib.ServerProxy


class ClientContainer(grok.Container):

    zope.interface.implements(gocept.lms.interfaces.IClientProvider)

    def add(self, client):
        self[client.id] = client


class Client(grok.Model):

    zope.interface.implements(gocept.lms.interfaces.IClient)

    id = None
    password = None
    contact_name = None
    contact_email = None
    callback = None

    urls = gocept.reference.ReferenceCollection(ensure_integrity=True)

    def __init__(self, id):
        self.id = id
        # This isn't a real set. It gets turned into something else by
        # gocept.relation.
        self.urls = set()

    def __cmp__(self, other):
        if isinstance(other, Client):
            return cmp(self.id, other.id)
        return -1

    def register_urls(self, urls):
        to_add = set(urls).difference(self.urls)
        for url in to_add:
            self.urls.add(url)
            url.clients.add(self)
            zope.event.notify(zope.lifecycleevent.ObjectModifiedEvent(url))

    def unregister_urls(self, urls):
        to_remove = set(urls).intersection(self.urls)
        for url in to_remove:
            self.urls.remove(url)
            url.clients.remove(self)
            zope.event.notify(zope.lifecycleevent.ObjectModifiedEvent(url))

    def iter_urls(self):
        return iter(self.urls)

    def remote_urls(self):
        """Get all URLs from the remote client."""
        return self.rpc.xmlrpc_getAllLinks(self.id, self.password)

    def notify_all(self):
        """Send notifications for all URLs of this client."""
        self.notify(self.urls)

    def notify_update(self):
        """Send notifications for all URLS of this client
        that had a status change since INotifications.last.

        """
        notifications = gocept.lms.interfaces.INotifications(self)
        url_query = zope.component.getUtility(hurry.query.interfaces.IQuery)
        urls = url_query.searchResults(
            Ge(('urls', 'last_state_change'), notifications.last) &
            AnyOf(('urls', 'clients'), [self]))
        if not urls:
            # No URLs have changed since last time. Nothing to do.
            return
        self.notify(urls)

    def notify(self, urls):
        """Send notifications about given urls."""
        log.info('Notifying client %s about URLs: %r' % (self.id, urls))
        notifications = [(url.url,
                          url.state or 'unknown', url.reason or 'unknown')
                         for url in urls]
        self.rpc.updateManyStates(self.id, self.password, notifications)

    @property
    def rpc(self):
        return ServerProxy(self.callback)


class NotificationsDisabled(object):
    """Event that signals that the notifications for a client where
    disabled.
    """

    def __init__(self, client):
        self.client = client


class Notifications(grok.Annotation, grok.Model):

    grok.provides(gocept.lms.interfaces.INotifications)
    grok.context(gocept.lms.interfaces.IClient)

    enabled = True
    last = datetime.datetime.min.replace(tzinfo=pytz.UTC)
    failed = 0

    def _p_resolveConflict(c, oldState, savedState, newState):
        # This usually happens with updating `last`. We're being
        # gracious about the other values.
        savedState['last'] = newState.get('last', c.last)
        savedState['enabled'] = (savedState.get('enabled', c.enabled) or
                                 newState.get('enabled', c.enabled))
        savedState['failed'] = min(savedState.get('failed', c.failed),
                                   newState.get('failed', c.failed))
        return savedState

    def notify(self):
        """Perform notifications on the recipient."""
        try:
            self.__parent__.notify_update()
        except Exception, e:
            log.exception(e)
            self.failed += 1
            if self.failed >= 20:
                log.warn('Disabling notifications for client %s due to too '
                         'many failed attempts.' % self.__parent__.id)
                self.enabled = False
                zope.event.notify(
                    gocept.lms.client.NotificationsDisabled(self.__parent__))
        else:
            self.failed = 0
            self.last = datetime.datetime.now(pytz.UTC)


class AddClient(grok.AddForm):
    """Register a new client."""

    grok.context(ClientContainer)
    grok.require('zope.ManageApplication')

    form_fields = grok.AutoFields(gocept.lms.interfaces.IClient)

    @grok.action('Add')
    def add(self, **data):
        client = Client(data['id'])
        grokcore.formlib.formlib.apply_data(client, self.form_fields, data)
        self.context.add(client)



# Copyright (c) 2008-2009 gocept gmbh & co. kg
# See also LICENSE.txt

import zope.interface
import zope.schema
import z3c.schema.email
import zc.sourcefactory.basic


STATE_OK = 'ok'
STATE_TEMPORARY = 'temporary unavailable'
STATE_UNAVAILABLE = 'unavailable'
STATE_CANT_CHECK = 'unsupported protocol'


class StateSource(zc.sourcefactory.basic.BasicSourceFactory):

    def getValues(self):
        return STATE_OK, STATE_TEMPORARY, STATE_UNAVAILABLE, STATE_CANT_CHECK


class IURLProvider(zope.interface.Interface):
    """URL container."""

    # XXX find a better name for this interface

    def add(url_string):
        """Create a new URL object and store it.

        Returns the new URL object. If an object representing the URL already
        exists, that object is returned.

        """

    def get_url(url_string):
        """Retrieve the URL object stored for the given URL.

        If no object for the given URL has been stored, None is returned.

        """


class IURL(zope.interface.Interface):

    url = zope.schema.URI(
        title=u'URL') # XXX make unique

    state = zope.schema.Choice(
        title=u'State',
        source=StateSource(),
        required=False)

    reason = zope.schema.TextLine(
        title=u'Reason for state',
        required=False)

    last_check = zope.schema.Datetime(
        title=u'Last checked',
        required=False)

    last_state_change = zope.schema.Datetime(
        title=u'State last changed',
        required=False)

    clients = zope.interface.Attribute(
        u'Iterable of all clients that registered this URL.')


class IClientProvider(zope.interface.Interface):
    """Client container."""

    # XXX find a better name for this interface

    def add(client):
        """Store a client object under its Id."""

    def __getitem__(id):
        """Retrieve a Client object by Id."""

    def get(id, default=None):
        """Retrieve a Client object by Id."""


class INotifications(zope.interface.Interface):
    """Manage notifications for a recipient."""

    enabled = zope.schema.Bool(
        title=u'Notify?',
        default=True)

    last = zope.schema.Datetime(
        title=u'Last notified')

    failed = zope.schema.Int(
        title=u'Number of failed notifications',
        default=0)

    def notify():
        """Perform notifications on the recipient."""


class IClient(zope.interface.Interface):

    id = zope.schema.TextLine(
        title=u'Client Id',
        description=u'Id used for authentication.') # XXX make unique

    password = zope.schema.Password(
        title=u'Password')

    contact_name = zope.schema.TextLine(
        title=u'Contact name',
        description=u'Real name of the person.')

    contact_email = z3c.schema.email.RFC822MailAddress(
        title=u'Contact email address')

    callback = zope.schema.URI(
        title=u'Callback URL')

    def iter_urls():
        """Return an iterable of the URLs registered for the client."""

    def remote_urls():
        """Get all URLs from the remote client."""

    def register_urls(urls):
        """Register URLs for client.

        urls: an iterable of IURL objects

        """

    def unregister_urls(urls):
        """Unregister URLs for client.

        urls: an iterable of IURL objects

        """

    def notify_all():
        """Send notifications for all URLs of this client."""

    def notify_update():
        """Send notifications for all URLS of this client
        that had a status change since INotifications.last.

        """

    def notify(urls):
        """Send notifications about given urls."""


class IThreadPool(zope.interface.Interface):
    """Thread pool for the checker."""

    limit = zope.schema.Int(
        title=u'Maximum number of concurrent active threads')

    active = zope.interface.Attribute(u'Set of active threads')

    available = zope.schema.Int(
        title=u'Number of currently available threads',
        readonly=True)


class ICheckerThread(zope.interface.Interface):

    url = zope.schema.URI(
        title=u'URL')

    state = zope.schema.Choice(
        title=u'State',
        source=StateSource(),
        required=False)

    reason = zope.schema.TextLine(
        title=u'Reason for state',
        required=False)


class ISchemeHandler(zope.interface.Interface):
    """Provides protocol-specific handling of a URL for one scheme."""

    def allow(url):
        """Determine whether to allow a URL to be registered or checked."""

    def check(url):
        """Check a URL and return a status tuple: (status, reason)"""

    def classify(url):
        """Determine the URL class."""


class ISynchronization(zope.interface.Interface):
    """Synchronize a client's database with the LMS's."""

    last = zope.schema.Datetime(
        title=u'Last notified')

    force = zope.schema.Bool(
        title=u'Force a synchronisation')

    def sync():
        """Perform the synchronization."""


class IStatistics(zope.interface.Interface):
    """Statistical information for munin."""

    thread_pool = zope.interface.Attribute('Number of threads currently used.')

    queue_size = zope.interface.Attribute('Number of items currently in the queue.')



# Copyright (c) 2008-2009 gocept gmbh & co. kg
# See also LICENSE.txt

import urlparse
import zope.component
import gocept.lms.interfaces


HOST_BLACKLIST = set(['localhost'])


def allow_host(host):
    return host not in HOST_BLACKLIST



# vim:fileencoding=utf-8
# Copyright (c) 2008-2009 gocept gmbh & co. kg
# See also LICENSE.txt
# $Id$
"""Base handler code."""

import grok
import gocept.lms.interfaces
import urlparse
import urllib


class AbstractURLHandler(object):
    """Abstract base handler implementation for URIs that are URLs.
    
    Subclasses need to implement `check`.

    """

    grok.provides(gocept.lms.interfaces.ISchemeHandler)

    def allow(self, url):
        scheme, netloc, path, params, query, fragment = urlparse.urlparse(url)
        if not netloc:
            return False
        host, port = urllib.splitport(netloc)
        return gocept.lms.filter.allow_host(host)

    def classify(self, url):
        scheme, netloc, path, params, query, fragment = urlparse.urlparse(url)
        if netloc:
            host, port = urllib.splitport(netloc)
            return host



# Copyright (c) 2008-2009 gocept gmbh & co. kg
# See also LICENSE.txt

import datetime
import gocept.lms.filter
import gocept.lms.interfaces
import grok
import logging
import pytz
import threading
import urllib
import zc.queue.interfaces
import zope.component
import zope.event
import zope.interface
import zope.lifecycleevent

log = logging.getLogger(__name__)

URL_INTERVAL = datetime.timedelta(seconds=60*60)
CLASS_INTERVAL = datetime.timedelta(seconds=1)

class_last = {}


def check_and_process():
    # Process first, so that the mainloop's sleep comes *after* spawning new
    # threads.
    process_finished()
    check()

    # Update statistical information. 
    # XXX This could probably be componentized using events.
    thread_pool = zope.component.getUtility(gocept.lms.interfaces.IThreadPool)
    statistics = zope.component.getUtility(gocept.lms.interfaces.IStatistics)
    statistics.thread_pool = len(thread_pool.active)


def check():
    """Pull URLs from the check queue and have them checked.

    Does so as long as there are threads available.

    """
    check_queue = zope.component.getUtility(zc.queue.interfaces.IQueue,
                                            name='check')
    thread_pool = zope.component.getUtility(gocept.lms.interfaces.IThreadPool)

    if not check_queue:
        # Shortcut. Also avoids logging.
        return

    log.info('checker: pulling urls from queue (%s entries)' % len(check_queue))

    now = datetime.datetime.now(pytz.UTC)
    mindate = datetime.datetime.min.replace(tzinfo=pytz.UTC)

    reschedule = set()

    while True:
        if not thread_pool.available:
            break
        if not check_queue:
            break

        url = check_queue.pull()
        handler = gocept.lms.interfaces.ISchemeHandler(url.url, None)

        if url.last_check > now - URL_INTERVAL:
            log.info('%s: ignored due to high frequency scheduling' % url.url)
            continue

        if handler is None:
            url.last_check = now
            zope.event.notify(zope.lifecycleevent.ObjectModifiedEvent(url))
            log.info('%s: ignored due to missing handler' % url.url)
            continue

        if not handler.allow(url.url):
            url.last_check = now
            zope.event.notify(zope.lifecycleevent.ObjectModifiedEvent(url))
            log.info('%s: ignored due to policy' % url.url)
            continue

        url_class = handler.classify(url.url)
        if class_last.get(url_class, mindate) > now - CLASS_INTERVAL:
            reschedule.add(url)
            continue

        log.info('%s: checking', url.url)
        class_last[url_class] = now
        thread = CheckerThread(url.url)
        thread.start()
        thread_pool.active.add(thread)

    for url in reschedule:
        check_queue.put(url)


def process_finished():
    thread_pool = zope.component.getUtility(gocept.lms.interfaces.IThreadPool)
    # Retrieve all finished threads and remove them from the active thread
    # list.
    finished = []
    for thread in list(thread_pool.active):
        if thread.isAlive():
            continue
        finished.append(thread)
        thread_pool.active.discard(thread)

    # Process all finished threads and write the status information back to
    # the URL objects.
    now = datetime.datetime.now(pytz.UTC)
    urls = zope.component.getUtility(gocept.lms.interfaces.IURLProvider)
    for thread in finished:
        url = urls.get_url(thread.url)
        log.info('%s: processing result' % url.url)
        log.info('%s: state %s -> %s' % (url.url, url.state, thread.state))
        url.last_check = now
        url.reason = thread.reason
        if url.state != thread.state:
            url.state = thread.state
            url.last_state_change = url.last_check
        zope.event.notify(zope.lifecycleevent.ObjectModifiedEvent(url))


class ThreadPool(grok.GlobalUtility):

    zope.interface.implements(gocept.lms.interfaces.IThreadPool)

    limit = 20

    def __init__(self):
        self.active = set()

    @property
    def available(self):
        remaining = self.limit - len(self.active)
        if remaining < 0:
            return 0
        return remaining


class CheckerThread(threading.Thread):

    zope.interface.implements(gocept.lms.interfaces.ICheckerThread)

    state = None
    reason = None

    def __init__(self, url):
        super(CheckerThread, self).__init__()
        self.url = url

    def run(self):
        type, remain = urllib.splittype(self.url)
        handler = zope.component.queryUtility(
            gocept.lms.interfaces.ISchemeHandler, name=type)
        if handler is None:
            self.state = gocept.lms.interfaces.STATE_CANT_CHECK
            self.reason = 'Can not check URLs for scheme: %s' % type
            return
        self.state, self.reason = handler.check(self.url)



# Copyright (c) 2008-2009 gocept gmbh & co. kg
# See also LICENSE.txt

import datetime
import logging
import pytz

import hurry.query
import hurry.query.interfaces
import zc.queue.interfaces
import zope.component

import gocept.lms.interfaces

log = logging.getLogger(__name__)

INTERVAL = datetime.timedelta(seconds=24*3600) # XXX make state-dependent
QUEUE_LIMIT = 1000
BACKOFF_TIME = 5*60


def schedule():
    check_queue = zope.component.getUtility(zc.queue.interfaces.IQueue,
                                            name='check')
    queue_len = len(check_queue)
    if queue_len > 1000:
        log.warn(
            'Scheduler found %s items in queue. Backing off for %s seconds'
            % (queue_len, BACKOFF_TIME))
        return BACKOFF_TIME
    url_query = zope.component.getUtility(hurry.query.interfaces.IQuery)
    reference_time = datetime.datetime.now(pytz.UTC) - INTERVAL
    urls = url_query.searchResults(
        hurry.query.Le(('urls', 'last_check'), reference_time))
    for url in urls:
        log.info('%s: scheduling' % url.url)
        check_queue.put(url)
    return



#!/usr/bin/env python
# vim:fileencoding=utf-8
# Copyright (c) 2008-2009 gocept gmbh & co. kg
# See also LICENSE.txt
# $Id$
"""LMS graph definitions for munin.

Symlink this script to /etc/munin/plugins/ and name it appropriately.

"""

from gocept.munin.client import SimpleGraph, main


class queuesize(SimpleGraph):
    key = name = 'queue-size'
    title = 'Checking queue size'
    category = 'LMS'


class threadpool(SimpleGraph):
    key = name = 'thread-pool'
    title = 'Active checker threads'
    category = 'LMS'

main()



import os.path
from setuptools import setup, find_packages


def read(*names):
    return open(
        os.path.join(os.path.dirname(__file__), *names)).read()


setup(name='gocept.lms',
      version='3.0a4',
      description="gocept Link Monitoring Server",
      long_description=(
          read('src', 'gocept', 'lms', 'README.txt')
          + '\n\n'
          + read('CHANGES.txt')
      ),
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: No Input/Output (Daemon)',
          'Framework :: Zope3',
          'Framework :: ZODB',
          'Intended Audience :: System Administrators',
          'License :: OSI Approved :: Zope Public License',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Topic :: Internet :: WWW/HTTP :: Site Management :: Link Checking',
          'Topic :: System :: Monitoring',
      ],
      keywords="zope lms link checking",
      author="gocept",
      author_email="mail@gocept.com",
      url="http://pypi.python.org/pypi/gocept.lms",
      license="ZPL 2.1",
      package_dir={'': 'src'},
      packages=find_packages('src'),
      include_package_data=True,
      zip_safe=False,
      namespace_packages=['gocept'],
      install_requires=[
          'decorator',
          'gocept.reference>=0.4dev',
          'gocept.runner',
          'grok',
          'hurry.query',
          'pytz',
          'setuptools',
          'z3c.schema',
          'z3c.testsetup',
          'zc.catalog',
          'zc.queue',
          'zc.relation',
          'zc.sourcefactory',
          'zope.app.wsgi',
          'zope.sendmail',
          'zope.testing',
      ],
      entry_points = dict(
        console_scripts =
          ['runscheduler = gocept.lms.runner:scheduler',
           'runchecker = gocept.lms.runner:checker',
           'runsyncer = gocept.lms.runner:syncer',
           'runnotifier = gocept.lms.runner:notifier'])
     )



# Copyright (c) 2008-2010 gocept gmbh & co. kg
# See also LICENSE.txt

import base64
import datetime
import gocept.lms.interfaces
import grok
import pytz
import urllib
import urlparse
import zope.component
import zope.interface
import zope.schema.interfaces


def prepare_urls(urls):
    """Take a list of URL strings and ensure they're valid URLs
    for the LMS.

    This method is a convenience to filter out invalid URL strings and
    may thus return a list that is shorter than the list passed in.

    """
    for url in urls:

        yield url


class URLContainer(grok.Container):

    zope.interface.implements(gocept.lms.interfaces.IURLProvider)

    def add(self, url_string):
        url_string = gocept.lms.interfaces.IURL['url'].fromUnicode(url_string)
        url = self.get_url(url_string)
        if url is None:
            url = URL(url_string)
            self[self.encode_url(url_string)] = url
        return url

    def add_multiple(self, urls):
        for url in urls:
            try:
                obj = self.add(url)
            except (UnicodeEncodeError, UnicodeDecodeError,
                    zope.schema.interfaces.InvalidURI):
                continue
            yield obj

    def get_url(self, url_string):
        return self.get(self.encode_url(url_string))

    @staticmethod
    def encode_url(url_string):
        return base64.urlsafe_b64encode(url_string)


class URL(grok.Model):

    zope.interface.implements(gocept.lms.interfaces.IURL)

    url = None
    state = None
    reason = None
    last_check = datetime.datetime.min.replace(tzinfo=pytz.UTC)
    last_state_change = datetime.datetime.min.replace(tzinfo=pytz.UTC)

    clients = gocept.reference.ReferenceCollection(ensure_integrity=True)

    def __init__(self, url_string):
        self.url = url_string
        self.clients = set()

    def __repr__(self):
        return "<%s.%s '%s'>" % (
            __name__, self.__class__.__name__, self.url)

    def __cmp__(self, other):
        if isinstance(other, URL):
            return cmp(self.url, other.url)
        return -1


class Purge(grok.View):

    grok.context(URL)

    def update(self):
        for client in self.context.clients:
            client.urls.remove(self.context)
        # After deletion we can't compute the application URL anymore so this
        # has to happen early.
        self.app_url = self.application_url()
        del self.context.__parent__[self.context.__name__]

    def render(self):
        self.redirect(self.app_url + '/urls')


@grok.adapter(str)
@grok.implementer(gocept.lms.interfaces.ISchemeHandler)
def get_handler(url):
    scheme, netloc, path, params, query, fragment = urlparse.urlparse(url)
    return zope.component.queryUtility(
        gocept.lms.interfaces.ISchemeHandler, name=scheme)



# vim:fileencoding=utf-8
# Copyright (c) 2008-2009 gocept gmbh & co. kg
# See also LICENSE.txt
#
# This installation serves as an initial generation as we didn't have
# any before this one, so it also checks for "very" old databases and
# performs initial migration if necessary.

import zope.app.zopeappgenerations
import gocept.lms.app
import zc.relation.interfaces


def evolve(context):
    root = zope.app.zopeappgenerations.getRootFolder(context)
    for candidate in root.values():
        if not isinstance(candidate, gocept.lms.app.LMS):
            continue
        # Those are very old LMS which still use zc.relation. We go
        # through all the registered URLs and check their client
        # relations and transform them to the new way (TM).
        old_site = zope.app.component.hooks.getSite()
        zope.app.component.hooks.setSite(candidate)
        try:
            urls = zope.component.getUtility(
                    gocept.lms.interfaces.IURLProvider)
            references = zope.component.getUtility(
                    zc.relation.interfaces.ICatalog, name='url_refs')
            for url in urls.values():
                url.clients = set()
                token = references.tokenizeValues([url], 'client_urls').next()
                for client in references.findRelations({'client_urls': token}):
                    url.clients.add(client)
                zope.event.notify(zope.lifecycleevent.ObjectModifiedEvent(url))
            # Uninstall zc.relation catalog
            assert candidate.getSiteManager().unregisterUtility(
                references, zc.relation.interfaces.ICatalog, name='url_refs')
            assert candidate.getSiteManager()['Catalog-2'] == references
            del candidate.getSiteManager()['Catalog-2']
        finally:
            zope.app.component.hooks.setSite(old_site)



# vim:fileencoding=utf-8
# Copyright (c) 2008-2009 gocept gmbh & co. kg
# See also LICENSE.txt
# $Id$
"""Checker implementation for the HTTP family."""

from gocept.lms.interfaces import STATE_OK, STATE_UNAVAILABLE, STATE_TEMPORARY
import gocept.lms.filter
import gocept.lms.interfaces
import grok
import httplib
import logging
import socket
import urllib
import urllib2
import urlparse

log = logging.getLogger(__name__)


http_opener = urllib2.OpenerDirector()
for handler in [urllib2.ProxyHandler,
                urllib2.HTTPHandler,
                urllib2.HTTPSHandler,
                urllib2.HTTPDefaultErrorHandler,
                urllib2.HTTPRedirectHandler,
                urllib2.HTTPErrorProcessor]:
    http_opener.add_handler(handler())


class HTTP(gocept.lms.handler.AbstractURLHandler,
           grok.GlobalUtility):

    grok.name('http')

    def check(self, url):
        request = urllib2.Request(url)
        request.add_header('user-agent', 'Bot/LMS (lms.gocept.com)')
        try:
            connection = http_opener.open(request)
            code = connection.code
        except urllib2.HTTPError, e:
            code = e.code
            reason = e.msg
        except urllib2.URLError, e:
            if isinstance(e.reason, socket.gaierror):
                reason = e.reason.args[1]
            else:
                reason = str(e.reason)
            return STATE_UNAVAILABLE, reason
        except httplib.BadStatusLine:
            return STATE_UNAVAILABLE, 'Presumably, the server closed the '\
                    'connection before sending a valid response.'
        else:
            connection.close()

        log.debug('%s: Got response code %s' % (url, code))
        status, status_msg = code_to_status.get(
            code, (STATE_UNAVAILABLE, 'Invalid response code'))

        if code == 599:
            # Testing support. The 5xx class causes an HTTPError to be  raised.
            status_msg = reason

        return status, status_msg


class HTTPS(HTTP):
    # Derive to register under different name
    grok.name('https')


code_to_status = {
    100: (STATE_OK, 'Continue'),
    101: (STATE_OK, 'Switching Protocols'),

    200: (STATE_OK, 'OK'),
    201: (STATE_OK, 'Created'),
    202: (STATE_OK, 'Accepted'),
    203: (STATE_OK, 'Non-Authoritative Information'),
    204: (STATE_OK, 'No Content'),
    205: (STATE_OK, 'Reset Content'),
    206: (STATE_OK, 'Partial Content'),

    300: (STATE_OK, 'Multiple Choices'),
    301: (STATE_OK, 'Moved Permanently'),
    302: (STATE_OK, 'Found'),
    303: (STATE_OK, 'See Other'),
    304: (STATE_OK, 'Not Modified'),
    305: (STATE_OK, 'Use Proxy'),
    306: (STATE_OK, '(Unused)'),
    307: (STATE_OK, 'Temporary Redirect'),

    400: (STATE_UNAVAILABLE, 'Bad Request'),
    401: (STATE_OK, 'Unauthorized'),
    402: (STATE_UNAVAILABLE, 'Payment required'),
    403: (STATE_UNAVAILABLE, 'Forbidden'),
    404: (STATE_UNAVAILABLE, 'Not Found'),
    405: (STATE_UNAVAILABLE, 'Method Not Allowed'),
    406: (STATE_UNAVAILABLE, 'Not Acceptable'),
    407: (STATE_UNAVAILABLE, 'Proxy Authentication Required'),
    408: (STATE_UNAVAILABLE, 'Request Timeout'),
    409: (STATE_UNAVAILABLE, 'Conflict'),
    410: (STATE_UNAVAILABLE, 'Gone'),
    411: (STATE_UNAVAILABLE, 'Length Required'),
    412: (STATE_UNAVAILABLE, 'Precondition failed'),
    413: (STATE_UNAVAILABLE, 'Request Entity Too Large'),
    414: (STATE_UNAVAILABLE, 'Request-URI Too Long'),
    415: (STATE_UNAVAILABLE, 'Unsupported Media Type'),
    416: (STATE_UNAVAILABLE, 'Request Range Not Satisfiable'),
    417: (STATE_UNAVAILABLE, 'Expectation Failed'),

    500: (STATE_TEMPORARY, 'Internal Server Error'),
    501: (STATE_UNAVAILABLE, 'Not Implemented'),
    502: (STATE_UNAVAILABLE, 'Bad Gateway'),
    503: (STATE_TEMPORARY, 'Service Unavailable'),
    504: (STATE_TEMPORARY, 'Gateway Timeout'),
    505: (STATE_UNAVAILABLE, 'HTTP Version Not Supported'),
    599: (STATE_UNAVAILABLE, 'Used internally by LMS'),
}
